//
//	tga.cpp - source file / freeware
//
//	David Henry - tfc_duke@club-internet.fr
//


#ifdef _WIN32
#define	WIN32_LEAN_AND_MEAN
#include	<windows.h>
#endif

#include	<GL/gl.h>
#include	<fstream>

#include	"texture.h"


// --------------------------------------------
// TGAHEADER - targa header.
// --------------------------------------------

#pragma pack(1)
typedef struct tagTGAHEADER
{
	unsigned char	id_lenght;			// taille du tagTGAHEADER
	unsigned char	color_map_type;		// 1 si l'image possde une palette
	unsigned char	image_type;			// type de compression de l'image

	short int		cm_first_entry;		// colormap_origin
	short int		cm_length;			// colormap_length
	unsigned char	cm_size;			// colormap_size

	short int		is_xorigin;			// coordonnes X du coin bas gauche de l'image
	short int		is_yorigin;			// coordonnes Y du coin bas gauche de l'image

	short int		is_width;			// largeur de l'image (en pixels)
	short int		is_height;			// hauteur de l'image (en pixels)

	unsigned char	is_pixel_depth;		// nombre de bits par pixel: 16, 24, 32
	unsigned char	is_image_descriptor;// 24 bits = 0x00; 32 bits = 0x80

} TGAHEADER, *PTGAHEADER;
#pragma pack(4)



// --------------------------------------------
// RGBTRIPLE - pixel 24 bits.
// --------------------------------------------
#ifndef _WIN32
typedef struct tagRGBTRIPLE			// rgbt
{
	unsigned char	rgbtBlue;		// bleu
	unsigned char	rgbtGreen;		// vert
	unsigned char	rgbtRed;		// rouge

} RGBTRIPLE, *PRGBTRIPLE;
#endif	// WIN32



// --------------------------------------------
// BGRAQUAD - pixel 32 bits.
// --------------------------------------------

typedef struct tagBGRAQUAD			// rgbt
{
	unsigned char	bgraBlue;		// bleu
	unsigned char	bgraGreen;		// vert
	unsigned char	bgraRed;		// rouge
	unsigned char	bgraAlpha;		// alpha

} BGRAQUAD, *PBGRAQUAD;



// --------------------------------------------------
// LoadFileTGA() - charge une image TGA. Si bFlipVert
// vaut 'true', l'image est retourne verticalement.
// formats accepts:
//     # BGR 16-24-32 bits
//     # BGR 16-24-32 bits (compression RLE)
//     # index couleur 8 bits
//     # index couleur 8 bits (compression RLE)
//     # grayscale 8-16 bits
//     # grayscale 8 bits (compression RLE)
//
// * Attention : les images en index couleur avec canal
// alpha peuvent provoquer des effets bizarres !
// * Attention : les images en dgrad de gris avec canal
// alpha et compression RLE ne sont pas supports !
// (grayscale 16 bits (compression RLE))
// --------------------------------------------------

bool LoadFileTGA( const std::string &szFileName, gl_texture_t **img, bool bFlipVert )
{
	std::ifstream	file;				// fichier
	TGAHEADER		*tgah = NULL;		// header TGA
	RGBTRIPLE		*palette = NULL;	// pointeur palette couleur
	char			*buffer = NULL;		// buffer stockant l'intgralit du fichier
	GLubyte			*ptr = NULL;		// pointeur donnes pixels
	int				row, col, i;		// variables de contrle



	/////////////////////////////////////////////////////
	// lit l'intgralit du fichier dans le buffer

	file.open( szFileName.c_str(), std::ios::in | std::ios::binary );

	if( file.fail() )
		return false;

	file.seekg( 0, std::ios::end );
	long flen = file.tellg();
	file.seekg( 0, std::ios::beg );

	buffer = new char[ flen ];
	file.read( buffer, flen );
	char *pBuff = buffer;

	file.close();

	/////////////////////////////////////////////////////


	// on lit le header
	tgah = (TGAHEADER *)pBuff;
	pBuff += sizeof( TGAHEADER );

	// cre un objet texture
	*img = new gl_texture_t;
	(*img)->width = tgah->is_width;
	(*img)->height = tgah->is_height;

	switch( tgah->image_type )
	{
		case 3:		// grayscale 8 bits
		case 11:	// grayscale 8 bits (compression RLE)
			if( (int)tgah->is_pixel_depth == 8 ) {
				(*img)->format = GL_LUMINANCE;
				(*img)->internalFormat = 1;
			}
			else { // 16 bits
				(*img)->format = GL_LUMINANCE_ALPHA;
				(*img)->internalFormat = 2;
			}

			break;

		case 1:		// index couleur 8 bits
		case 2:		// BGR 16-24-32 bits
		case 9:		// index couleur 8 bits (compression RLE)
		case 10:	// BGR 16-24-32 bits (compression RLE)
			// les images 8 et 16 bits seront converties en 24 bits
			if( (int)tgah->is_pixel_depth <= 24 ) {
				(*img)->format = GL_RGB;
				(*img)->internalFormat = 3;
			}
			else { // 32 bits
				(*img)->format = GL_RGBA;
				(*img)->internalFormat = 4;
			}

			break;
	}

	// allocation de mmoire pour stocker les donnes pixels
	(*img)->texels = new GLubyte[ (*img)->width * (*img)->height * (*img)->internalFormat ];
	ptr = &(*img)->texels[0];

	// on dplace le pointeur fichier au dbut des donnes pixels
	if( tgah->id_lenght )
		pBuff = buffer + (tgah->id_lenght * sizeof( unsigned char ));


	/////////////////////////////////////////////////////
	// lecture de la palette

	if( tgah->color_map_type )
	{
		// les images 16, 24 et 32 bits ne possdent pas de palettes,
		// ainsi que les images en dgrad de gris

		// on ajuste le pointeur de la palette sur la zone mmoire
		// correspondante dans 'buffer'
		palette = (RGBTRIPLE *)pBuff;
		pBuff += tgah->cm_length * (tgah->cm_size >> 3) * sizeof( unsigned char );
	}

	/////////////////////////////////////////////////////


	/////////////////////////////////////////////////////
	// lecture des donnes suivant le type de
	// compression et le nombre de bits par pixel
	/////////////////////////////////////////////////////

	switch( tgah->image_type )
	{
		case 0:
			// pas de donnes image
			break;

		case 1:
		{
// COLOR-MAPPED BGR 8 BITS
			for( row = (*img)->height - 1; row >= 0; row-- )
			{
				if( bFlipVert )
					ptr = &(*img)->texels[ row * (*img)->width * (*img)->internalFormat ];

				for( col = 0; col < (int)(*img)->width; col++, ptr += 3 )
				{
					// lecture du pixel en cours
					unsigned char color = *((unsigned char *)(pBuff++));

					// conversion pixel index couleur bgr en pixel rgb 24 bits
					ptr[0] = (GLubyte)palette[ color ].rgbtRed;		// b->r
					ptr[1] = (GLubyte)palette[ color ].rgbtGreen;	// g->g
					ptr[2] = (GLubyte)palette[ color ].rgbtBlue;	// r->b
				}
			}

			break;
		}

		case 2:
		{
			for( row = (*img)->height - 1; row >= 0; row-- )
			{
				if( bFlipVert )
					ptr = &(*img)->texels[ row * (*img)->width * (*img)->internalFormat ];

				for( col = 0; col < (int)(*img)->width; col++ )
				{
					switch( tgah->is_pixel_depth )
					{
						case 16:
						{
// TRUE-COLOR BGR 16 BITS
							// lecture du pixel en cours
							unsigned short color = *((unsigned short *)pBuff);
							pBuff += sizeof( short );

							// convertion pixel bgr 16 bits en rgb 24 bits
							ptr[0] = (GLubyte)(((color & 0x7C00) >> 10) << 3);	// b->r
							ptr[1] = (GLubyte)(((color & 0x03E0) >>  5) << 3);	// g->g
							ptr[2] = (GLubyte)(((color & 0x001F) >>  0) << 3);	// r->b
							ptr += 3;
							break;
						}

						case 24:
						{
// TRUE-COLOR BGR 24 BITS
							// lecture du pixel 24 bits et conversion bgr->rgb
							RGBTRIPLE *pix = (RGBTRIPLE *)pBuff;
							pBuff += sizeof( RGBTRIPLE );

							ptr[0] = (GLubyte)pix->rgbtRed;
							ptr[1] = (GLubyte)pix->rgbtGreen;
							ptr[2] = (GLubyte)pix->rgbtBlue;
							ptr += 3;
							break;
						}

						case 32:
						{
// TRUE-COLOR BGR 32 BITS
							// lecture du pixel 32 bits et conversion bgr->rgb
							BGRAQUAD *pix = (BGRAQUAD *)pBuff;
							pBuff += sizeof( BGRAQUAD );

							ptr[0] = (GLubyte)pix->bgraRed;
							ptr[1] = (GLubyte)pix->bgraGreen;
							ptr[2] = (GLubyte)pix->bgraBlue;
							ptr[3] = (GLubyte)pix->bgraAlpha;
							ptr += 4;
							break;
						}
					}
				}
			}

			break;
		}

		case 3:
		{
// 8 BITS GREYSCALE
			for( row = (*img)->height - 1; row >= 0; row-- )
			{
				if( bFlipVert )
					ptr = &(*img)->texels[ row * (*img)->width * (*img)->internalFormat ];

				for( col = 0; col < (int)((*img)->width) * (*img)->internalFormat; col++, ptr++ )
					*ptr = *((GLubyte *)(pBuff++));
			}

			break;
		}

		case 9:
		{
// RLE COLOR-MAPPED BGR 8 BITS
			unsigned char	packetHeader, packetSize, i;

			for( row = (*img)->height - 1; row >= 0; row-- )
			{
				if( bFlipVert )
					ptr = &(*img)->texels[ row * (*img)->width * (*img)->internalFormat ];

				for( col = 0; col < (int)(*img)->width; /* rien */ )
				{
					packetHeader	= *((unsigned char *)(pBuff++));
					packetSize		= 1 + (packetHeader & 0x7f);

					if( packetHeader & 0x80 )
					{
						// run-length packet

						// lecture du pixel en cours
						unsigned char color = *((unsigned char *)(pBuff++));

						// conversion pixel index couleur bgr en pixel rgb 24 bits
						for( i = 0; i < packetSize; i++, ptr += 3, col++ )
						{
							ptr[0] = (GLubyte)palette[ color ].rgbtRed;		// b->r
							ptr[1] = (GLubyte)palette[ color ].rgbtGreen;	// g->g
							ptr[2] = (GLubyte)palette[ color ].rgbtBlue;	// r->b
						}
					}
					else
					{
						// non run-length packet

						for( i = 0; i < packetSize; i++, ptr += 3, col++ )
						{
							// lecture du pixel en cours
							unsigned char color = *((unsigned char *)(pBuff++));

							// conversion pixel index couleur bgr en pixel rgb 24 bits
							ptr[0] = (GLubyte)palette[ color ].rgbtRed;		// b->r
							ptr[1] = (GLubyte)palette[ color ].rgbtGreen;	// g->g
							ptr[2] = (GLubyte)palette[ color ].rgbtBlue;	// r->b
						}
					}
				}
			}

			break;
		}

		case 10:
		{
			unsigned char	packetHeader, packetSize;

			for( row = (*img)->height - 1; row >= 0; row-- )
			{
				if( bFlipVert )
					ptr = &(*img)->texels[ row * (*img)->width * (*img)->internalFormat ];

				for( col = 0; col < (int)(*img)->width; /* rien */ )
				{
					packetHeader	= *((unsigned char *)(pBuff++));
					packetSize		= 1 + (packetHeader & 0x7f);

					if( packetHeader & 0x80 )
					{
						// run-length packet
						switch( tgah->is_pixel_depth )
						{
							case 16:
							{
// RLE TRUE-COLOR BGR 16 BITS
								// lecture du pixel en cours
								unsigned short color = *((unsigned short *)pBuff);
								pBuff += sizeof( short );

								// convertion pixel bgr 16 bits en rgb 24 bits
								for( i = 0; i < packetSize; i++, ptr += 3, col++ )
								{
									ptr[0] = (GLubyte)(((color & 0x7C00) >> 10) << 3);	// b->r
									ptr[1] = (GLubyte)(((color & 0x03E0) >>  5) << 3);	// g->g
									ptr[2] = (GLubyte)(((color & 0x001F) >>  0) << 3);	// r->b
								}

								break;
							}

							case 24:
							{
// RLE TRUE-COLOR BGR 24 BITS
								// lecture du pixel 24 bits et conversion bgr->rgb
								RGBTRIPLE *pix = (RGBTRIPLE *)pBuff;
								pBuff += sizeof( RGBTRIPLE );

								for( i = 0; i < packetSize; i++, ptr += 3, col++ )
								{
									ptr[0] = (GLubyte)pix->rgbtRed;
									ptr[1] = (GLubyte)pix->rgbtGreen;
									ptr[2] = (GLubyte)pix->rgbtBlue;
								}

								break;
							}

							case 32:
							{
// RLE TRUE-COLOR BGR 32 BITS
								// lecture du pixel 32 bits et conversion bgr->rgb
								BGRAQUAD *pix = (BGRAQUAD *)pBuff;
								pBuff += sizeof( BGRAQUAD );

								for( i = 0; i < packetSize; i++, ptr += 4, col++ )
								{
									ptr[0] = (GLubyte)pix->bgraRed;
									ptr[1] = (GLubyte)pix->bgraGreen;
									ptr[2] = (GLubyte)pix->bgraBlue;
									ptr[3] = (GLubyte)pix->bgraAlpha;
								}

								break;
							}
						}
					}
					else
					{
						// non run-length packet

						for( i = 0; i < packetSize; i++, col++ )
						{
							switch( tgah->is_pixel_depth )
							{
								case 16:
								{
// RLE TRUE-COLOR BGR 16 BITS
									// lecture du pixel en cours
									unsigned short color = *((unsigned short *)pBuff);
									pBuff += sizeof( short );

									// convertion pixel bgr 16 bits en rgb 24 bits
									ptr[0] = (GLubyte)(((color & 0x7C00) >> 10) << 3);	// b->r
									ptr[1] = (GLubyte)(((color & 0x03E0) >>  5) << 3);	// g->g
									ptr[2] = (GLubyte)(((color & 0x001F) >>  0) << 3);	// r->b
									ptr += 3;
									break;
								}

								case 24:
								{
// RLE TRUE-COLOR BGR 24 BITS
									// lecture du pixel 24 bits et conversion bgr->rgb
									RGBTRIPLE *pix = (RGBTRIPLE *)pBuff;
									pBuff += sizeof( RGBTRIPLE );

									ptr[0] = (GLubyte)pix->rgbtRed;
									ptr[1] = (GLubyte)pix->rgbtGreen;
									ptr[2] = (GLubyte)pix->rgbtBlue;
									ptr += 3;
									break;
								}

								case 32:
								{
// RLE TRUE-COLOR BGR 32 BITS
									// lecture du pixel 32 bits et conversion bgr->rgb
									BGRAQUAD *pix = (BGRAQUAD *)pBuff;
									pBuff += sizeof( BGRAQUAD );

									ptr[0] = (GLubyte)pix->bgraRed;
									ptr[1] = (GLubyte)pix->bgraGreen;
									ptr[2] = (GLubyte)pix->bgraBlue;
									ptr[3] = (GLubyte)pix->bgraAlpha;
									ptr += 4;
									break;
								}
							}
						}
					}
				}
			}

			break;
		}

		case 11:
		{
// RLE 8 BITS GREYSCALE
			// pas de support pour les images compresses RLE en
			// dgrad de gris avec canal alpha
			if( (int)tgah->is_pixel_depth > 8 )
			{
				delete [] buffer;
				delete *img;
				return false;
			}

			unsigned char	packetHeader, packetSize, i;

			for( row = (*img)->height - 1; row >= 0; row-- )
			{
				if( bFlipVert )
					ptr = &(*img)->texels[ row * (*img)->width * (*img)->internalFormat ];

				for( col = 0; col < (int)((*img)->width) * (*img)->internalFormat; /* rien */ )
				{
					packetHeader	= *((unsigned char *)(pBuff++));
					packetSize		= 1 + (packetHeader & 0x7f);

					if( packetHeader & 0x80 )
					{
						// run-length packet

						// lecture du pixel en cours
						GLubyte color = *((GLubyte *)(pBuff++));

						for( i = 0; i < packetSize; i++, ptr++, col++ )
							*ptr = color;
					}
					else
					{
						// non run-length packet

						for( i = 0; i < packetSize; i++, ptr++, col++ )
							*ptr = *((GLubyte *)(pBuff++));
					}
				}
			}

			break;
		}

		default:
		{
			// format inconnu ou non trat
			delete [] (*img)->texels;
			delete [] buffer;
			return false;
		}
	}

	delete [] buffer;
	return true;
}
